(*

An implementation of environments, which is how the evaluator stores the
mapping from variables to their contents and how the typechecker stores the
mapping from variables to their types.  The Assoc_List.iterate
calls are good examples of how first-class functions can be used.

*)

structure Env = struct

(* A mapping from strings to values of type 'a.  Both environments and
type environments are implemented as ProtoEnvironments (see below),
allowing them to share much of their functionality. *)
type 'a ProtoEnvironment = (string, 'a) Assoc_List.assoc_list

(* empty_environment is the initial environment with no bound variables *)
val empty_environment:'a ProtoEnvironment = Assoc_List.empty

(* add_binding adds a binding to a ProtoEnvironment,
returning a new extended ProtoEnvironment. *)
fun add_binding(env:'a ProtoEnvironment, name:string,
                value:'a):'a ProtoEnvironment =
  Assoc_List.store(env, name, value)

(* new_binding creates a new ProtoEnvironment containing a single binding
of name->value *)
fun new_binding(name:string, value:'a):'a ProtoEnvironment =
    add_binding(empty_environment, name, value)

(* NotFound is the exception raised if lookups or removes fail *)
exception NotFound = Assoc_List.NotFound

(* lookup_binding takes a ProtoEnvironment and a name, and returns the
value bound to that name in the ProtoEnvironment, or raises NotFound if the
name was unbound *)
fun lookup_binding(env:'a ProtoEnvironment, name:string):'a =
    Assoc_List.fetch(env, name)

(* DuplicateVar is the exception raised if a duplicate variable is found
   while trying to merge environments that should have disjoint domains. *)
exception DuplicateVar of string

(* combine_envs takes two ProtoEnvironments and adds them together to form
a new, merged environment containing all the bindings of the two
combined ProtoEnvironments.  if any name is defined in both ProtoEnviroments,
the binding in env2 takes precedence (shadows) that in env1 *)
fun combine_envs(env1:'a ProtoEnvironment,
                 env2:'a ProtoEnvironment):'a ProtoEnvironment =
    let fun combine_one((name:string, value:'a), env):'a ProtoEnvironment =
	        add_binding(env, name, value)
    in Assoc_List.iterate(combine_one, env1, env2) end

(* combine_envs_no_dups is like combine_envs, but it ensures that the two
   environments to combine have disjoint domains. it raises DuplicateVar
   if a duplicate is found.  *)
fun combine_envs_no_dups(env1:'a ProtoEnvironment,
			 env2:'a ProtoEnvironment):'a ProtoEnvironment =
    let fun combine_one((name:string, value:'a), env):'a ProtoEnvironment =
	let val _ = lookup_binding(env, name) 
	in raise DuplicateVar(name) end
        handle NotFound => add_binding(env, name, value)
    in Assoc_List.iterate(combine_one, env1, env2) end

(* protoEnvToString converts a ProtoEnvironment to a printable string.
   separator is a string to use between the name and its associated value.
   valToString is a function to convert values to strings.
   notice the use of higher-order functions! *)
fun protoEnvToString(env:'a ProtoEnvironment, separator:string,
		     valToString:'a -> string):string =
    let fun oneToString((name:string, value:'a), prefix:string):string =
                prefix ^ "val " ^ name ^ separator ^ valToString(value) ^"\n"
    in Assoc_List.iterate(oneToString, "", env) end

  
local
  open Ast
  open AstPrint
  open Value
  open ValuePrint
in

(* ENVIRONMENTS *)
    
(* an environment stores a list of bindings, one for each variable
that's bound, mapping each variable to its current Value *)
type Environment = Value ProtoEnvironment

(* TYPE ENVIRONMENTS *)

(* a type environment stores a list of bindings, one for each variable
that's bound, mapping each variable to its Type *)
type TypeEnvironment = Type ProtoEnvironment


(* envAndTypeEnvToString prints environment and type environment bindings in a
   unified manner *)
fun envAndTypeEnvToString(env:Environment, tenv:TypeEnvironment):string =
    let fun oneToString((name:string, value:Value), prefix:string):string =
	prefix ^ "val " ^ name ^ " = " ^ valueToString(value) ^
	" : " ^ typeToString(lookup_binding(tenv, name)) ^ "\n"
    in Assoc_List.iterate(oneToString, "", env) end

end

end
