(* Lexer for MiniML language. *)

(** Stuff to get yacc to work right **)
type pos=int
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult = (svalue,pos) token
    
exception LexerError of string;

val pos = ref 1
val nesting_depth = ref 0
fun error(x) = 
    raise LexerError("line " ^ Int.toString(!pos)  ^ ": error: " ^ x)
fun eof() = let val p = !pos in (pos := 0; Tokens.EOF(p,p)) end

fun parse_int(s:string):int =
    (case Int.fromString(s) of
        SOME(i) => i
      | NONE => error("illegal integer literal"))
    handle Overflow => error("integer literal too large")

fun parse_hexint(s:string):int =
    let val (sign,s) = if String.sub(s, 0) = #"~" 
		       then (~1, String.extract(s, 1, NONE))
		       else ( 1, s)
        val s = String.extract(s, 2, NONE)  (* drop 0x *)
        val digits = explode(s)
        fun as_int(digit) =
	    if Char.isDigit(digit) then ord(digit) - ord(#"0")
	    else ord(Char.toLower(digit)) - ord(#"a") + 10
        fun convert(nil) = (0, 1)
          | convert(digit::digits) =
	        let val (res, base) = convert(digits)
	        in (as_int(digit) * base + res, base * 16) end
        val (i, _) = convert(digits)
    in sign * i end
    handle Overflow => error("integer literal too large")

fun parse_word(s:string):word =
    let val s = String.extract(s, 2, NONE)  (* drop 0w *)
        val digits = explode(s)
        fun as_word(digit) = Word.fromInt(ord(digit) - ord(#"0"))
        fun convert(nil) = (0w0, 0w1)
          | convert(digit::digits) =
	        let val (res, base) = convert(digits)
	        in (as_word(digit) * base + res, base * 0w10) end
        val (w, _) = convert(digits)
    in w end
    handle Overflow => error("word literal too large")

fun parse_hexword(s:string):word =
    (case Word.fromString(s) of
        SOME(w) => w
      | NONE => error("illegal word literal"))
    handle Overflow => error("word literal too large")

fun parse_real(s:string):real =
    case Real.fromString(s) of
        SOME(r) => r
      | NONE => error("illegal real literal")

fun parse_escapes(s:string):string =
    case String.fromString(s) of
        SOME(s') => if String.toString(s') = s then s'
		    else error("illegal escape sequence")
      | NONE => error("illegal escape sequence")

fun parse_string(s:string):string =
    let val no_quotes = String.substring(s,1,String.size(s)-2)
	val no_escapes = parse_escapes(no_quotes)
    in no_escapes end

fun parse_char(s:string):char =
    let val without_hash = String.extract(s,1,NONE)
        val no_escapes = parse_string(without_hash)
    in if String.size(no_escapes) = 1 then String.sub(no_escapes, 0)
       else error("character literal should have a single character") end

%%

%s COMMENT;
%header (functor MiniMLLexFun(structure Tokens: MiniML_TOKENS));

ws=[\ \t];
digit=[0-9];
hexdigit=[0-9a-fA-F];
int=~?{digit}+;
hexint=~?0x{hexdigit}+;
real={int}(\.[0-9]+([eE]{int})?|[eE]{int});
alpha=[A-Za-z'];
id={alpha}[A-Za-z'0-9_]*;
operator=[-!@#$%^&*+=/:<>?\\~`|];
str=\"[^\"]*\";
comment_substr=[^\*\(\n]*;

%%

"(*"		=> (if (!nesting_depth = 0) then 
			(YYBEGIN(COMMENT);
			nesting_depth:= (!nesting_depth+1))
		    else
			nesting_depth:= (!nesting_depth+1);		
		    lex());

\n 		=> (pos := 1 + !pos; lex());
<COMMENT> "*"	=> (lex());
<COMMENT> "("	=> (lex());
<COMMENT> {comment_substr} => (lex());
<COMMENT> "*)"		=> (nesting_depth:= (!nesting_depth-1);
			    if !nesting_depth = 0 then YYBEGIN(INITIAL) 
			    else ();
			    lex());
<INITIAL> "*)"		=> (error("unmatched close comment"));

<INITIAL> {ws}+		=> (lex());

<INITIAL> {int}		=> (
    Tokens.INT(parse_int(yytext),!pos,!pos));

<INITIAL> {hexint}	=> (
    Tokens.INT(parse_hexint(yytext),!pos,!pos));

<INITIAL> 0w{int}	=> (
    Tokens.WORD(parse_word(yytext),!pos,!pos));

<INITIAL> 0wx{hexint}	=> (
    Tokens.WORD(parse_hexword(yytext),!pos,!pos));

<INITIAL> {real}	=> (
    Tokens.REAL(parse_real(yytext),!pos,!pos));

<INITIAL> {str}		=> (
    Tokens.STR(parse_string(yytext),!pos,!pos));

<INITIAL> #{str}	=> (
    Tokens.CHAR(parse_char(yytext),!pos,!pos));

<INITIAL> "abstype"	=> (Tokens.ABSTYPE(!pos,!pos));
<INITIAL> "and"		=> (Tokens.AND(!pos,!pos));
<INITIAL> "andalso"	=> (Tokens.ANDALSO(!pos,!pos));
<INITIAL> "as"		=> (Tokens.AS(!pos,!pos));
<INITIAL> "case"	=> (Tokens.CASE(!pos,!pos));
<INITIAL> "datatype"	=> (Tokens.DATATYPE(!pos,!pos));
<INITIAL> "do"		=> (Tokens.DO(!pos,!pos));
<INITIAL> "else"	=> (Tokens.ELSE(!pos,!pos));
<INITIAL> "end"		=> (Tokens.END(!pos,!pos));
<INITIAL> "eqtype"	=> (Tokens.EQTYPE(!pos,!pos));
<INITIAL> "exception"	=> (Tokens.EXCEPTION(!pos,!pos));
<INITIAL> "fn"		=> (Tokens.FN(!pos,!pos));
<INITIAL> "fun"		=> (Tokens.FUN(!pos,!pos));
<INITIAL> "functor"	=> (Tokens.FUNCTOR(!pos,!pos));
<INITIAL> "handle"	=> (Tokens.HANDLE(!pos,!pos));
<INITIAL> "if"		=> (Tokens.IF(!pos,!pos));
<INITIAL> "in"		=> (Tokens.IN(!pos,!pos));
<INITIAL> "include"	=> (Tokens.INCLUDE(!pos,!pos));
<INITIAL> "infix"	=> (Tokens.INFIX(!pos,!pos));
<INITIAL> "infixr"	=> (Tokens.INFIXR(!pos,!pos));
<INITIAL> "let"		=> (Tokens.LET(!pos,!pos));
<INITIAL> "local"	=> (Tokens.LOCAL(!pos,!pos));
<INITIAL> "nonfix"	=> (Tokens.NONFIX(!pos,!pos));
<INITIAL> "of"		=> (Tokens.OF(!pos,!pos));
<INITIAL> "op"		=> (Tokens.OP(!pos,!pos));
<INITIAL> "open"	=> (Tokens.OPEN(!pos,!pos));
<INITIAL> "orelse"	=> (Tokens.ORELSE(!pos,!pos));
<INITIAL> "raise"	=> (Tokens.RAISE(!pos,!pos));
<INITIAL> "rec"		=> (Tokens.REC(!pos,!pos));
<INITIAL> "sharing"	=> (Tokens.SHARING(!pos,!pos));
<INITIAL> "sig"		=> (Tokens.SIG(!pos,!pos));
<INITIAL> "signature"	=> (Tokens.SIGNATURE(!pos,!pos));
<INITIAL> "struct"	=> (Tokens.STRUCT(!pos,!pos));
<INITIAL> "structure"	=> (Tokens.STRUCTURE(!pos,!pos));
<INITIAL> "then"	=> (Tokens.THEN(!pos,!pos));
<INITIAL> "type"	=> (Tokens.TYPE(!pos,!pos));
<INITIAL> "val"		=> (Tokens.VAL(!pos,!pos));
<INITIAL> "with"	=> (Tokens.WITH(!pos,!pos));
<INITIAL> "withtype"	=> (Tokens.WITHTYPE(!pos,!pos));
<INITIAL> "where"	=> (Tokens.WHERE(!pos,!pos));
<INITIAL> "while"	=> (Tokens.WHILE(!pos,!pos));

<INITIAL> "("	=> (Tokens.LPAREN(!pos,!pos));
<INITIAL> ")"	=> (Tokens.RPAREN(!pos,!pos));
<INITIAL> "{"	=> (Tokens.LBRACE(!pos,!pos));
<INITIAL> "}"	=> (Tokens.RBRACE(!pos,!pos));
<INITIAL> "["	=> (Tokens.LBRACKET(!pos,!pos));
<INITIAL> "]"	=> (Tokens.RBRACKET(!pos,!pos));
<INITIAL> ","	=> (Tokens.COMMA(!pos,!pos));
<INITIAL> ":"	=> (Tokens.COLON(!pos,!pos)); 
<INITIAL> ";"	=> (Tokens.SEMI_COLON(!pos,!pos));
<INITIAL> "..."	=> (Tokens.ELIPSIS(!pos,!pos));
<INITIAL> "_"	=> (Tokens.UNDERSCORE(!pos,!pos));
<INITIAL> "|"	=> (Tokens.BAR(!pos,!pos));
<INITIAL> "="	=> (Tokens.EQUALS(!pos,!pos));
<INITIAL> "=>"	=> (Tokens.EQUALS_GT(!pos,!pos));
<INITIAL> "->"	=> (Tokens.ARROW(!pos,!pos));
<INITIAL> "#"	=> (Tokens.HASH(!pos,!pos));
<INITIAL> "."	=> (Tokens.DOT(!pos,!pos));
<INITIAL> ":>"	=> (Tokens.COLON_GT(!pos,!pos));

<INITIAL> "not"		=> (Tokens.NOT(!pos,!pos));
<INITIAL> "div"		=> (Tokens.DIV(!pos,!pos));
<INITIAL> "mod"		=> (Tokens.MOD(!pos,!pos));
<INITIAL> "true"	=> (Tokens.BOOL(true,!pos,!pos));
<INITIAL> "false"	=> (Tokens.BOOL(false,!pos,!pos));
<INITIAL> "nil"		=> (Tokens.NIL(!pos,!pos));

<INITIAL> "~"	=> (Tokens.TILDE(!pos,!pos));
<INITIAL> "+"	=> (Tokens.PLUS(!pos,!pos));
<INITIAL> "-"	=> (Tokens.MINUS(!pos,!pos));
<INITIAL> "*"	=> (Tokens.TIMES(!pos,!pos));
<INITIAL> "/"	=> (Tokens.DIVIDE(!pos,!pos));
<INITIAL> "<"	=> (Tokens.LT(!pos,!pos));
<INITIAL> "<="	=> (Tokens.LEQ(!pos,!pos));
<INITIAL> ">="	=> (Tokens.GEQ(!pos,!pos));
<INITIAL> ">"	=> (Tokens.GT(!pos,!pos));
<INITIAL> "<>"	=> (Tokens.LT_GT(!pos,!pos));
<INITIAL> "::"	=> (Tokens.COLON_COLON(!pos,!pos));
<INITIAL> "^"	=> (Tokens.CARAT(!pos,!pos));

<INITIAL> "int"		=> (Tokens.INT_TYPE(!pos,!pos));
<INITIAL> "string"	=> (Tokens.STR_TYPE(!pos,!pos));
<INITIAL> "bool"	=> (Tokens.BOOL_TYPE(!pos,!pos));
<INITIAL> "unit"	=> (Tokens.UNIT_TYPE(!pos,!pos));
<INITIAL> "list"	=> (Tokens.LIST_TYPE(!pos,!pos));


<INITIAL> {id}		=> (if String.sub(yytext,0) = #"'"
			    then Tokens.TYVAR(yytext,!pos,!pos)
			    else Tokens.ID(yytext,!pos,!pos));

<INITIAL> {operator}+	=> (Tokens.OPERATOR(yytext,!pos,!pos));

.	=> (error("bad character " ^ yytext));
