(*

This file contains functions related to handling explicitly quantified
types.

*)

structure Type_Quant = struct

local open Ast in

(* some helper operations for treating a list of type variables as a set *)
fun isMember elem = List.exists (fn x => x=elem)
fun addElem(var, lst) = if isMember var lst then lst else var::lst
fun union(lst1,lst2) = foldl addElem lst2 lst1

(* return the list of all names (strings) of type variables that appear free
   in the given type *)
fun getFreeVars(typ:Type):string list =
    let fun freeVarsHelper(IntType, boundVars, currFreeVars) = currFreeVars
	  | freeVarsHelper(StrType, boundVars, currFreeVars) = currFreeVars
	  | freeVarsHelper(BoolType, boundVars, currFreeVars) = currFreeVars
	  | freeVarsHelper(UnitType, boundVars, currFreeVars) = currFreeVars
	    (* add a type variable to the list of free variables if the
	       variable is not currently bound *)
	  | freeVarsHelper(TypeVar(name), boundVars, currFreeVars) =
	      if isMember name boundVars then currFreeVars
	      else addElem(name, currFreeVars)

	    (* check the components of a tuple type for free variables from
	       left to right, updating the current set of free vars
	       as we go *)
	  | freeVarsHelper(TupleType(tList), boundVars, currFreeVars) =
	      foldl (fn(typ, freeVars) =>
		      freeVarsHelper(typ, boundVars, freeVars))
	            currFreeVars tList
		    
	  | freeVarsHelper(ListType(typ), boundVars, currFreeVars) =
	      freeVarsHelper(typ, boundVars, currFreeVars)
	  | freeVarsHelper(FunctionType{argType, resType}, boundVars,
			   currFreeVars) =
	      let val currFreeVars' = freeVarsHelper(argType, boundVars,
						     currFreeVars)
	      in freeVarsHelper(resType, boundVars, currFreeVars') end
	 
	   (* add all the quantified variables to the list of bound vars
	      before recursing on the underlying type *)
	  | freeVarsHelper(ForallType{typevars,typ}, boundVars,
			   currFreeVars) =
	      let val boundVars' = union(typevars, boundVars)
	      in freeVarsHelper(typ, boundVars', currFreeVars) end
    in freeVarsHelper(typ, nil, nil) end

end

end
