(*

 An implementation of type substitutions, which are used in typechecking
 calls to (polymorphic) functions.  A type substitution is a mapping from
 type variables to types.  It represents the particular instantiation of the
 function's domain type to the given argument type.

*)

structure Type_Subst = struct

local
  open Ast
  open Type_Quant
in

(* a mapping from the names of type variables to the types that they should
   be instantiated with.  *)
type TypeSubstitution = (string, Type) Assoc_List.assoc_list

(* empty_type_subst is the mapping with an empty domain. *)
val empty_type_subst:TypeSubstitution = Assoc_List.empty

(* NotFound is the exception raised if lookup_type_var fails *)
exception NotFound = Assoc_List.NotFound

(* lookup_type_var takes a TypeSubstitution and a name, and returns
the type bound to that name in the TypeSubstitution, or raises
NotFound if the name was unbound *)
fun lookup_type_var(subst:TypeSubstitution, tyvar:string):Type =
    Assoc_List.fetch(subst, tyvar)

(* add_type_var adds a pair to a TypeSubstitution,
 returning a new extended TypeSubstitution. *)
fun add_type_var(subst:TypeSubstitution, name:string,
		 typ:Type):TypeSubstitution =
    Assoc_List.store(subst, name, typ)

(* apply a substitution to a type and return the instantiated type.
   we allow the substitution to be partial:  it need not be the case that all
   type vars in the given type are in the domain of the substitution.
   make sure not to substitute for a bound type variable! *)
fun apply_type_subst(subst:TypeSubstitution, t:Type):Type =
    let fun apply_subst_helper(IntType, boundVars):Type = IntType
	  | apply_subst_helper(StrType, boundVars):Type = StrType
	  | apply_subst_helper(BoolType, boundVars):Type = BoolType
	  | apply_subst_helper(UnitType, boundVars):Type = UnitType

	    (* if the given type variable is bound, then leave it as is.
	       otherwise substitute for the given type variable as directed by
	       the given type substitution.  if the variable is not in the
	       substitution, also leave it as is. *)
	  | apply_subst_helper(typ as TypeVar(name), boundVars):Type =
	      if isMember name boundVars then typ
	      else (lookup_type_var(subst, name) handle NotFound => typ)
		
	  | apply_subst_helper(TupleType(tlist), boundVars):Type =
		TupleType(map
			  (fn t => apply_subst_helper(t, boundVars)) tlist)
	  | apply_subst_helper(ListType(typ), boundVars):Type =
		ListType(apply_subst_helper(typ, boundVars))
	  | apply_subst_helper(FunctionType{argType,resType}, boundVars) =
		FunctionType{argType=apply_subst_helper(argType, boundVars),
			     resType=apply_subst_helper(resType, boundVars)}

	    (* add the newly-bound type variables to the list of bound vars
	       for the recursive call *)
	  | apply_subst_helper(ForallType{typevars, typ}, boundVars):Type =
		ForallType{typevars=typevars,
			   typ=apply_subst_helper(typ,
						  union(typevars,boundVars))}
    in apply_subst_helper(t, nil) end

end

end
