(*

This file defines the way to do type inference and checking of all the
different kinds of ASTs.

*)


structure Typecheck = struct

local
  open Ast
  open AstPrint
  open Env
  open Unify
in


(* if a case is initially unimplemented, but needs to be implemented
for successful completion of the project, then it raises the
ImplementMe exception, passing a description of the unimplemented
feature as the argument *)

exception ImplementMe of string


(* if any typecheck functions encounter a type error, then they raise
the TypeError exception, with a string explaining the error *)

exception TypeError of string


(* a short-hand *)
type UnifyState = Unify.T


(* NOTE: many of the operations below make repeated calls to Unify
operations, each producing a new unify_state that must be passed on to
the next call to a Unify operation.  In other words, there should be
one unify_state "active" at a time, and each call to a Unify operation
typically produces a new unify_state that should take over the active
role.  This is known as "single-threading," since there is a single
thread of active unify_states through the program.  To avoid
accidentally refering to an old unify_state, I adopt the convention of
binding new unify_states to names that shadow ("redeclare") the
previous unify_state, so that I can never accidentally refer to it,
and so that I don't have to make up new names for each of the
unify_states that are produced.  This is a useful programming
technique when making a series of "updates" to a side-effect-free data
structure.  See typecheckPattern for example usages. *)


(********** HELPERS ********)

(* opType returns the type of the corresponding operator.  it may be
   polymorphic, so it must be instantiated before being applied to an
   argument. *)
	
fun opType("~") = FunctionType{argType=IntType, resType=IntType}
  | opType("not") = FunctionType{argType=BoolType, resType=BoolType}

  | opType("+") = FunctionType{argType=TupleType[IntType,IntType],
			       resType=IntType}
  | opType("-") = FunctionType{argType=TupleType[IntType,IntType],
			       resType=IntType}
  | opType("*") = FunctionType{argType=TupleType[IntType,IntType],
			       resType=IntType}
  | opType("div") = FunctionType{argType=TupleType[IntType,IntType],
			         resType=IntType}
  | opType("mod") = FunctionType{argType=TupleType[IntType,IntType],
			         resType=IntType}
    
  | opType("<") = FunctionType{argType=TupleType[IntType,IntType],
			       resType=BoolType}
  | opType("<=") = FunctionType{argType=TupleType[IntType,IntType],
			        resType=BoolType}
  | opType(">=") = FunctionType{argType=TupleType[IntType,IntType],
			        resType=BoolType}
  | opType(">") = FunctionType{argType=TupleType[IntType,IntType],
			       resType=BoolType}

  | opType("^") = FunctionType{argType=TupleType[StrType,  StrType],
			       resType=StrType}

(* for now, don't handle overloaded string comparison operators
  | opType("<") = FunctionType{argType=TupleType[StrType,  StrType],
			       resType=BoolType}
  | opType("<=") = FunctionType{argType=TupleType[StrType,  StrType],
			        resType=BoolType}
  | opType(">=") = FunctionType{argType=TupleType[StrType,  StrType],
			        resType=BoolType}
  | opType(">") = FunctionType{argType=TupleType[StrType,  StrType],
			       resType=BoolType}
*)

  | opType("::") =
      (* 'a * 'a list -> 'a list *)
      ForallType{typevars=["'a"],
		 typ=FunctionType{argType=TupleType[TypeVar("'a"),
					            ListType(TypeVar("'a"))],
			          resType=ListType(TypeVar("'a"))}}

  | opType("=") =
      (* ''a * ''a -> bool *)
      ForallType{typevars=["''a"],
		 typ=FunctionType{argType=TupleType[TypeVar("''a"),
					            TypeVar("''a")],
			          resType=BoolType}}
  | opType("<>") = opType("=")

  | opType(name) =
      (* an internal error *)
      raise TypeError("unexpected operator " ^ name)


(* typecheck an application of a function or operator to an argument
   type, returning the type of the result of the application *)

fun typecheckApp(e:Expr, fnType:Type, argType:Type, unify_state:UnifyState)
		:Type * UnifyState =
      raise ImplementMe("typecheckApp")
      (* to typecheck an application: *)
	(* instantiate function type, if necessary *)
	(* construct "type" of call: a function type from the argument
	   type to a fresh type variable standing for the result *)
	(* unify the instantiated function type with the call type (we
	   do it via unification because we don't know that the
	   function's type is actually a FunctionType; e.g., it might
	   still be an unconstrained type variable) *)
(* for debugging; uncomment this code if you want to see the unification
   environment at each call, assuming val bindings for names like
   instantiatedFnType and resType
        val _ = (print("at call " ^ exprToString(e) ^ ":\n");
		 print("  fun type: " ^ typeToString(fnType) ^ "\n");
		 print("  instantiated fun type: " ^ 
		       typeToString(instantiatedFnType) ^ "\n");
		 print("  arg type: " ^ typeToString(argType) ^ "\n");
		 print("  result type: " ^ typeToString(resType) ^ "\n");
		 print("  unification state:\n" ^
		       toStringIndenting unify_state "    ");
		 print("  resolved fun type: " ^ 
		       typeToString(resolve unify_state instantiatedFnType) ^
		       "\n");
		 print("  resolved arg type: " ^ 
		       typeToString(resolve unify_state argType) ^ "\n");
		 print("  resolved result type: " ^ 
		       typeToString(resolve unify_state resType) ^ "\n");
		 print("\n"))
*)
      (* return call's result type *)


(* typecheck a pattern, returning its type and a set of type bindings for
   variables in the pattern *)

fun typecheckPattern(VarPattern{name,typ=SOME t}, unify_state:UnifyState)
		    :Type * TypeEnvironment * UnifyState =
      (* if a type is given for a variable pattern, then use it *)
      (t, new_binding(name, t), unify_state)
  | typecheckPattern(VarPattern{name,typ=NONE}, unify_state) =
      (* if no type is given for a variable pattern, then make up a
         fresh type variable for it *)
      let val (t:Type, unify_state) = fresh_type_var unify_state
      in (t, new_binding(name, t), unify_state) end

  | typecheckPattern(WildcardPattern{typ=SOME t}, unify_state) =
      (* if a type is given for a wildcard pattern, then use it *)
      (t, empty_environment, unify_state)
  | typecheckPattern(WildcardPattern{typ=NONE}, unify_state)
		    :Type * TypeEnvironment * UnifyState =
      (* if no type is given for a wildcard pattern, then make up a
         fresh type variable for it *)
      let val (t:Type, unify_state) = fresh_type_var unify_state
      in (t, empty_environment, unify_state) end

  | typecheckPattern(IntPattern(i), unify_state) =
      (IntType, empty_environment, unify_state)
  | typecheckPattern(StringPattern(s), unify_state) =
      (StrType, empty_environment, unify_state)
  | typecheckPattern(BoolPattern(i), unify_state) =
      (BoolType, empty_environment, unify_state)

  | typecheckPattern(p as NilPattern{typ}, unify_state) =
      let
        (* create a fresh type variable to stand for the element type *)
	val (elemType:Type, unify_state) = fresh_type_var unify_state

	(* create the type of the result empty list *)
	val listType:Type = ListType(elemType)

	(* if an explicit type is given, then unify it with the list type *)
	val unify_state =
	  case typ of
	      NONE => unify_state
	    | SOME(lt:Type) => unify unify_state lt listType
				     (fn()=>patternToString(p))

      in (listType, empty_environment, unify_state) end

  | typecheckPattern(TuplePattern[], unify_state) =
      (UnitType, empty_environment, unify_state)

  | typecheckPattern(p as TuplePattern(pats), unify_state) =
      (* to typecheck a tuple pattern, typecheck each of its component
	 patterns, producing a type and type environment for each.
	 combine the type environments, ensuring there are no
	 duplicate variables in their domains, and create a tuple type
	 of the component types. *)
      (let val (types:Type list, bindings:TypeEnvironment, unify_state) =
             foldr
	       (fn(pat:Pattern, (prevTypes, prevBindings, unify_state))=>
		 let val (patType:Type,bindings:TypeEnvironment,unify_state) =
		       typecheckPattern(pat, unify_state)
		 in (patType::prevTypes, 
		     combine_envs_no_dups(prevBindings, bindings),
		     unify_state) end)
	       (nil, empty_environment, unify_state) pats
       in (TupleType(types), bindings, unify_state) end

       handle DuplicateVar(name) =>
	 raise TypeError("duplicate variable " ^ name ^ " in pattern " ^
			 patternToString(p)))

  | typecheckPattern(p as ConsPattern{head, tail}, unify_state) =
      let
        (* typecheck the head and tail patterns *)
        val (headType:Type, headBindings:TypeEnvironment, unify_state) =
	  typecheckPattern(head, unify_state)
        val (tailType:Type, tailBindings:TypeEnvironment, unify_state) =
	  typecheckPattern(tail, unify_state)

        (* combine the bindings in the two patterns, reporting an
	   error if there are duplicate bindings *)
	val bindings:TypeEnvironment =
	  combine_envs_no_dups(headBindings, tailBindings)
            handle DuplicateVar(name) =>
	      raise TypeError("duplicate variable " ^ name ^ " in pattern " ^
		              patternToString(p))

	(* construct a list type from the type of the head pattern,
	   and unify it with the type of the tail *)
	val listType:Type = ListType(headType)
	val unify_state = unify unify_state listType tailType
				(fn()=>patternToString(p))

      (* return the list type as the type of the pattern, plus the
         bindings and the new unification state *)
      in (listType, bindings, unify_state) end
					 

(******* MUTUALLY RECURSIVE TYPECHECK FUNCTIONS OVER DECLS AND EXPRS *******)

(* typecheckDecl typechecks a declaration to produce a set of new
   bindings and an updated unification state.
   tenv is a type environment mapping the variables in scope to their types.
   unify_state is the state of the unifier before typechecking the
   declaration. *)

fun typecheckDecl(d as ValDecl{lhs:Pattern,rhs:Expr},
		  tenv:TypeEnvironment,
		  unify_state:UnifyState):TypeEnvironment*UnifyState =
      raise ImplementMe("typecheckDecl on ValDecls")
        (* typecheck the pattern, producing a type for the pattern and
	   type bindings for variables in the pattern *)
        (* typecheck the rhs expr, producing a type *)
	(* unify the types of the pattern and the rhs *)
	(* generalize any free, generalizable type variables in the
	   bindings, replacing them with explicit quantification *)
      (* return the new, generalized bindings, plus the new
	 unification state *)
		
  | typecheckDecl(d as FunDecl(cases),
		  tenv:TypeEnvironment,
		  unify_state:UnifyState):TypeEnvironment*UnifyState =
      raise ImplementMe("typecheckDecl on FunDecls")
        (* get the function's name from the first case *)
	(* make up fresh type variables for the argument and result types *)
	(* augment the environment with a binding for the function, to
	   handle recursive calls (monomorphically) *)
	(* iterate over each of the cases, typechecking them in the
	   environment including the function type, and accumulating
	   unification constraints [use the typecheckCase helper below] *)
        (* create a binding for the function *)
        (* generalize any free, generalizable type variables in the
	   binding, replacing them with explicit quantification *)
      (* return the new, generalized binding, plus the new
	 unification state *)


(* typecheck a function case *)

and typecheckCase(cs as {name:string, formal:Pattern,
			 resType:Type option, body:Expr},
		  tenv:TypeEnvironment,
		  unify_state:UnifyState,
		  funName:string, arg_type:Type, res_type:Type):UnifyState =
  (* check that the case has the expected name *)
  if name <> funName
    then raise TypeError("case " ^ name ^ " doesn't match function name " ^
			 funName)
  else
    let
      (* typecheck the pattern, producing a type for the pattern and
	 type bindings for variables in the pattern *)
      val (patType:Type, bindings:TypeEnvironment, unify_state) =
	typecheckPattern(formal, unify_state)

      (* unify the pattern's type with the function's argument type *)
      val unify_state = unify unify_state patType arg_type
				      (fn()=>caseToString(cs))

      (* if specified, unify the function's specified return type with
	 its assumed return type *)
      val unify_state =
        case resType of
	    SOME(specifiedResType) =>
	      unify unify_state specifiedResType res_type
	            (fn()=>caseToString(cs))
	  | NONE => unify_state

      (* typecheck the body in the current type environment augmented
	 with the bindings from the formal pattern *)
      val (bodyType:Type, unify_state) =
	typecheckExpr(body, combine_envs(tenv, bindings), unify_state)

      (* unify the body's type with the function's return type *)
      val unify_state = unify unify_state bodyType res_type
			      (fn()=>caseToString(cs))

    (* return the final unification state *)
    in unify_state end


(* typecheckExpr typechecks an expression in a type environment to
   produce a result Type. *)

and typecheckExpr(VarExpr(var_name:string),
		  tenv:TypeEnvironment, 
		  unify_state:UnifyState):Type*UnifyState =
      raise ImplementMe("typecheckExpr of VarExprs")
        (* look up the variable's binding in the current type
	   environment, raising an error if the variable isn't bound *)
        (* the variable might be polymorphic; if so, instantiate it by
	   replacing any quantified type variables with fresh type
	   variables *)
      (* return the instantiated type *)

  (* to typecheck a constant, just return the corresponding type *)
  | typecheckExpr(IntExpr(v), tenv, unify_state) =
      raise ImplementMe("typecheckExpr of IntExprs")
  | typecheckExpr(StringExpr(v), tenv, unify_state) =
      raise ImplementMe("typecheckExpr of StringExprs")
  | typecheckExpr(BoolExpr(v), tenv, unify_state) =
      raise ImplementMe("typecheckExpr of BoolExprs")
	
  | typecheckExpr(e as NilExpr{typ}, tenv, unify_state) =
      raise ImplementMe("typecheckExpr of NilExprs")
        (* create a fresh type variable to stand for the element type *)
	(* create the type of the result empty list *)
	(* if an explicit type is given, then unify it with the list type *)
	(* return the list type *)

  | typecheckExpr(TupleExpr(nil), tenv, unify_state) =
      raise ImplementMe("typecheckExpr of TupleExprs(nil)")
      (* the empty tuple has type unit *)

  | typecheckExpr(TupleExpr(elems), tenv, unify_state) =
      raise ImplementMe("typecheckExpr of TupleExprs(non-nil)")
      (* to typecheck a tuple expression, typecheck each of the
         subexpressions, then construct the tuple type *)
    
  | typecheckExpr(e as UnopExpr{operator:string, arg:Expr},
		  tenv, unify_state) =
      raise ImplementMe("typecheckExpr of UnopExprs")
        (* typecheck argument expression *)
	(* look up type of operator *)
      (* typecheck the application of the operator to the argument *)

  | typecheckExpr(e as BinopExpr{arg1:Expr, operator:string, arg2:Expr},
		  tenv, unify_state) =
      raise ImplementMe("typecheckExpr of BinopExprs")
        (* typecheck argument expressions *)
	(* look up type of operator *)
	(* construct the pair of argument expressions *)
      (* typecheck the application of the operator to the argument pair *)

  | typecheckExpr(e as CallExpr{function:string, arg:Expr},
		  tenv, unify_state) =
      raise ImplementMe("typecheckExpr of CallExprs")
        (* look up function name's binding in the current environment,
	   raising an error if it's not bound *)
        (* typecheck argument expression *)
      (* typecheck the application of the function to the argument *)
	    
  | typecheckExpr(e as IfExpr{testExpr, thenExpr, elseExpr},
		  tenv, unify_state) =
      raise ImplementMe("typecheckExpr of IfExprs")
        (* typecheck the test expressions *)
	(* unify the test expression's type with the bool type *)
        (* typecheck the then and else expressions *)
	(* unify their types *)
      (* return the then (or else) type *)
    
  | typecheckExpr(LetExpr{decls:Decl list, body:Expr},
		  tenv, unify_state) =
      raise ImplementMe("typecheckExpr of LetExprs")
      (* iterate through each of the declarations,
         typechecking them in the context of the current environment
	 augmented with the bindings introduced by each of the previous
	 declarations.  then typecheck the body expression in the fully
	 augmented environment and return the result type of the body
	 typecheck. *)


(********** MAIN TYPECHECK FUNCTION ********)

(* typecheckTopLevel typechecks the top-level ast in a fresh
   unification context *)

fun typecheckTopLevel(TopLevelDecl(decl), tenv):TypeEnvironment =
     (let
        (* typecheck the declaration in a fresh unification state *)
        val (bindings:TypeEnvironment, unify_state) =
          typecheckDecl(decl, tenv, Unify.new)

        (* resolve any constrained variables in the bindings *)
        val resolvedBindings:TypeEnvironment =
	  resolveTypeEnv unify_state bindings

      (* return the resolved bindings, throwing away the unification state *)
      in resolvedBindings end

      handle UnificationError(msg) => raise TypeError(msg))

  | typecheckTopLevel(TopLevelExpr(expr), tenv) =
      (* treat a top-level expression as if it were "val it = expr" *)
      let val it_decl = ValDecl{lhs=VarPattern{name="it",typ=NONE}, rhs=expr}
      in typecheckTopLevel(TopLevelDecl(it_decl), tenv) end


(* typecheck takes the top-level decl or expr, and the top-level
type environment (i.e., the bindings of global variables to their types),
and then typechecks the top-level ast to produce a set of new bindings
(possibly of "it"), which is returned to the caller *)

fun typecheck(ast:TopLevel, tenv:TypeEnvironment):TypeEnvironment =
  typecheckTopLevel(ast, tenv)

end

end
