(*

This file defines a module supporting type unification operations.

*)


signature UNIFY = sig

  (* the type of the unification state *)
  type T

  (* the initial unification state, representing no constraints *)
  val new: T

  (* create a fresh, unconstrained type variable, along with the
     unification state to use for all future operations relating to this
     fresh type variable *)
  val fresh_type_var: T -> Ast.Type * T

  (* constrain two types to be equal, i.e., unify them, in the context
     of a unification state.  return the new unification state that
     includes the new constraint.  if the constraint cannot be satisfied,
     then UnificationError is raised; the string argument of the
     UnificationError exception is computed by invoking the argument
     function.  (the string typically describes the expression causing the 
     unification constraint.) *)
  val unify: T -> Ast.Type -> Ast.Type -> (unit -> string) -> T
  exception UnificationError of string

  (* take a type, and return one where all type variables have been
     replaced with whatever they've been unified with, if anything,
     recursively.  (this function is used whenever constraint solving
     is complete, and the result of constraint solving is desired.) *)
  val resolve: T -> Ast.Type -> Ast.Type

  (* resolve each of the types in a type environment *)
  val resolveTypeEnv: T -> Env.TypeEnvironment -> Env.TypeEnvironment

  (* take a type and instantiate any quantified type variables with
     fresh type variables *)
  val instantiateType: T -> Ast.Type -> Ast.Type * T

  (* replace any types containing free, generalizable type variables
     in the first argument type environment with explicitly quantified types
     in the result type environment.  the second argument type
     environment is the one enclosing the type environment being
     generalized, to identify which free type variables are also free
     in the enclosing type environment; such type variables are not
     generalizable.  (in addition to handling the issue of
     non-generalizable type variables internally, it also accounts for
     any constraints that may have been applied to the types in either
     type environment.) *)
  val generalizeTypeEnv: T -> Env.TypeEnvironment -> Env.TypeEnvironment
			   -> Env.TypeEnvironment

  (* return a string print-out of the unification state *)
  val toString: T -> string

  (* similar, but prefix each line of the result with the argument
     string *)
  val toStringIndenting: T -> string -> string

end


structure Unify :> UNIFY = struct

local
  open Ast
  open AstPrint
  open Env
  open Type_Quant
  open Type_Subst
in

  type T = {
    (* the mapping from type variable names to the Types that they've
       been constrained to equal *)
    map:TypeSubstitution,

    (* a number used to generate the name of the next fresh type variable *)
    type_var_counter:int}

  exception UnificationError of string

  val new = {map=empty_type_subst, type_var_counter=0}

  fun fresh_type_var_of_kind {map,type_var_counter} isEqType =
    let val prefix:string = if isEqType then "''" else "'"
        val typeName:string = prefix ^ Int.toString(type_var_counter)
    in (TypeVar(typeName), {map=map, type_var_counter=type_var_counter+1}) end

  fun fresh_type_var unify_state =
    fresh_type_var_of_kind unify_state false

  fun fresh_eq_type_var unify_state =
    fresh_type_var_of_kind unify_state true

  fun resolve (unify_state:T) typ =
    let fun resolve_helper IntType boundNames = IntType
          | resolve_helper StrType boundNames = StrType
          | resolve_helper BoolType boundNames = BoolType
          | resolve_helper UnitType boundNames = UnitType

          | resolve_helper (t as TypeVar(name)) boundNames =
              (* if this type variable is bound in an enclosing forall
		 type, then don't change it *)
              if isMember name boundNames then t else
	      (let 
                  (* figure out what name is mapped to, if anything *)
                  val binding = lookup_type_var(#map unify_state, name)
	       in
	          (* if it was bound, then recursively resolve the target *)
	          resolve_helper binding boundNames
	       end
	         (* if it was not bound, then leave it alone *)
	         handle NotFound => t)

          | resolve_helper (TupleType(ts)) boundNames =
              TupleType(map (fn(t)=>resolve_helper t boundNames) ts)

          | resolve_helper (ListType(typ)) boundNames =
              ListType(resolve_helper typ boundNames)

          | resolve_helper (FunctionType{argType,resType}) boundNames =
	      let val resolvedArg = resolve_helper argType boundNames
	          val resolvedRes = resolve_helper resType boundNames
	      in FunctionType{argType=resolvedArg, resType=resolvedRes} end

          | resolve_helper (t as ForallType{typevars,typ}) boundNames =
              (* we need to ignore any constraints on type variables
		 matching any of the quantified types, since any
		 occurrences of the type variables in the body type
		 refer to the quantified type variables, not to any
		 type variables of the same name in the enclosing
		 unification context. *)
              let val resolvedTyp = resolve_helper typ (typevars@boundNames)
	      in ForallType{typevars=typevars, typ=resolvedTyp} end

    in resolve_helper typ nil end


  fun resolveTypeEnv unify_state tenv =
        Assoc_List.iterate(
          (fn((name:string,typ:Type), prevBindings:TypeEnvironment) =>
             add_binding(prevBindings, name, resolve unify_state typ)),
          empty_environment, tenv)


  (* a private helper function that adds a new TypeVar->Type mapping
     to the unification state.  TypeVar must not previously be
     constrained, and Type must be fully resolved.  if TypeVar is an
     equality type var, then Type must only have equality types in it.
     raises UnificationError if the TypeVar appears in Type, e.g.
     T = (T->T), since then a circularity would be created. *)
  fun add_mapping {map,type_var_counter} name typ msg_fn =
    if isMember name (getFreeVars typ) then
      raise UnificationError("circularity: " ^ name ^ " cannot match " ^
			     typeToString(typ) ^ " at " ^ msg_fn())
    else
      {map = add_type_var(map, name, typ), type_var_counter = type_var_counter}


  (* a private helper function that returns whether its argument is
     the name of an equality type variable, e.g., if its name starts
     with two single quotes *)
  fun is_equality_type(name:string):bool =
    String.sub(name, 0) = #"'" andalso String.sub(name, 1) = #"'"


  (* a private helper function that takes a fully resolved type and
     rewrites it as an equality type, if necessary, replacing any
     unconstrained type variables with equality type variables.
     additional unification constraints may be added.  the resulting
     type is fully resolved.  raises UnificationError if the argument
     type isn't an equality type *)
  fun make_eq_type unify_state typ msg_fn :Type * T =
    let fun helper unify_state IntType boundNames = (IntType, unify_state)
          | helper unify_state StrType boundNames = (StrType, unify_state)
          | helper unify_state BoolType boundNames = (BoolType, unify_state)
          | helper unify_state UnitType boundNames = (UnitType, unify_state)

          | helper unify_state (t as TypeVar(name)) boundNames =
              (* if this type variable is bound in an enclosing forall
		 type, then don't change it *)
              if isMember name boundNames then (t, unify_state)

	      (* if it's already an eq type, then nothing more to do *)
	      else if is_equality_type(name) then (t, unify_state)

	      else
	        (* otherwise, first see if it's been replaced with an
		   equality type already, by resolving it *)
	        let val resolved:Type = resolve unify_state t
		in
	          if t <> resolved then
	            (* it's been replaced; return its replacement (which
		       should be an equality type variable) *)
		    (resolved, unify_state)

	          else
	            (* we need to constrain this type variable to be an
		       equality type. *)
		    let
		      (* make a fresh equality type to replace it with. *)
		      val (eqT:Type, unify_state) =
			fresh_eq_type_var unify_state

		      (* add a mapping from the unconstrained type var
		         to the new equality type var *)
		      val unify_state = add_mapping unify_state name eqT msg_fn

		    (* then return the equality type *)
		    in (eqT, unify_state) end
	        end

          | helper unify_state (TupleType(ts)) boundNames =
              let val (eqTs:Type list, unify_state) =
		    foldr
		      (fn(t:Type, (prevEqTs:Type list, unify_state))=>
		        let val (eqT:Type, unify_state) =
			      helper unify_state t boundNames
			in (eqT::prevEqTs, unify_state) end)
		      (nil, unify_state) ts
              in (TupleType(eqTs), unify_state) end

          | helper unify_state (ListType(typ)) boundNames =
              let val (eqTyp:Type, unify_state) =
		    helper unify_state typ boundNames
	      in (ListType(eqTyp), unify_state) end

          | helper unify_state (t as FunctionType _) boundNames =
              raise UnificationError(typeToString(typ) ^
				     " isn't an equality type in " ^ msg_fn())

          | helper unify_state (t as ForallType{typevars,typ}) boundNames =
              (* we need to ignore any type variables
		 matching any of the quantified types, since any
		 occurrences of the type variables in the body type
		 refer to the quantified type variables, not to any
		 type variables of the same name in the enclosing
		 unification context. *)
              let val (eqTyp:Type, unify_state) =
		    helper unify_state typ (typevars@boundNames)
	      in (ForallType{typevars=typevars, typ=eqTyp}, unify_state) end

    in helper unify_state typ nil end


  fun unify unify_state IntType IntType msg_fn = unify_state
    | unify unify_state StrType StrType msg_fn = unify_state
    | unify unify_state BoolType BoolType msg_fn = unify_state
    | unify unify_state UnitType UnitType msg_fn = unify_state

    | unify unify_state (t1 as TypeVar(name1)) (t2 as TypeVar(name2)) msg_fn =
        let 
            (* first, figure out what each of the type vars has been
	       constrained to, if anything *)
            val resolved1 = resolve unify_state t1
            val resolved2 = resolve unify_state t2
	in
	    if resolved1 = resolved2 then
	      (* they're already constrained to be equal *)
	      unify_state
            else if t1 = resolved1 then
	      (* t1 hasn't been mapped to anything before *)
              if t2 = resolved2 then
	        (* t2 hasn't been mapped to anything before, either *)
	        if is_equality_type(name1) then
		  (* if t1 is an equality type, then map t2 to t1,
		     since t1 is potentially more constrained than t2 *)
	          add_mapping unify_state name2 resolved1 msg_fn
		else
		  (* otherwise, map t1 to t2, since t2 is potentially
		     more constrained than t1 *)
	          add_mapping unify_state name1 resolved2 msg_fn
	      else
	        (* t2 has been mapped to something *)
		if is_equality_type(name1) then
		  (* ensure that the target is also an equality type *)
		  let val (resolvedEqType2:Type, unify_state) =
		        make_eq_type unify_state resolved2
				     (fn()=> typeToString(resolved2) ^ 
				             " in " ^ msg_fn())
		  (* then map t1 to it *)
		  in add_mapping unify_state name1 resolvedEqType2 msg_fn end
                else
		  (* map t1 to it *)
		  add_mapping unify_state name1 resolved2 msg_fn
	    else if t2 = resolved2 then
	      (* t2 hasn't been mapped to anything before, but t1 has *)
	      if is_equality_type(name2) then
		(* ensure that the target is also an equality type *)
		let val (resolvedEqType1:Type, unify_state) =
		      make_eq_type unify_state resolved1
				   (fn()=> typeToString(resolved2) ^ 
				           " in " ^ msg_fn())
		(* then map t2 to it *)
		in add_mapping unify_state name2 resolvedEqType1 msg_fn end
              else
		(* map t2 to it *)
	        add_mapping unify_state name2 resolved1 msg_fn
	    else
	      (* they both have been constrained to something, and
		 those somethings are different.  so recursively unify
		 those somethings *)
	      unify unify_state resolved1 resolved2 msg_fn
        end

    | unify unify_state (t1 as TypeVar(name1)) t2 msg_fn =
        let 
            (* first, figure out what the arguments have been
	       constrained to, if anything *)
            val resolved1 = resolve unify_state t1
            val resolved2 = resolve unify_state t2
	in
	    if resolved1 = resolved2 then
	      (* they're already constrained to be equal *)
	      unify_state
            else if t1 = resolved1 then
	      (* t1 hasn't been mapped to anything before *)
	      if is_equality_type(name1) then
		(* ensure that the target is also an equality type *)
		let val (resolvedEqType2:Type, unify_state) =
		      make_eq_type unify_state resolved2
				   (fn()=> typeToString(resolved2) ^ 
				           " in " ^ msg_fn())
		(* then map t1 to it *)
		in add_mapping unify_state name1 resolvedEqType2 msg_fn end
              else
		(* map t1 to it *)
		add_mapping unify_state name1 resolved2 msg_fn
	    else
	      (* it has been constrained to something else already.
	         so recursively unify them *)
	      unify unify_state resolved1 resolved2 msg_fn
        end

    | unify unify_state t1 (t2 as TypeVar(_)) msg_fn =
        (* commute the arguments and retry *)
	unify unify_state t2 t1 msg_fn

    | unify unify_state (t1 as TupleType(tlist1))
			(t2 as TupleType(tlist2)) msg_fn =
        if length(tlist1) <> length(tlist2) then
          raise UnificationError(typeToString(t1) ^ " doesn't match " ^ 
			         typeToString(t2) ^ " at " ^ msg_fn())
        else
	  ListPair.foldl
	    (fn(type1,type2,unify_state)=>unify unify_state type1 type2 msg_fn)
            unify_state (tlist1, tlist2)

    | unify unify_state (ListType(type1)) (ListType(type2)) msg_fn =
	unify unify_state type1 type2 msg_fn

    | unify unify_state (FunctionType{argType=aType1, resType=rType1})
                        (FunctionType{argType=aType2, resType=rType2}) msg_fn =
	let val unify_state = unify unify_state aType1 aType2 msg_fn
	    val unify_state = unify unify_state rType1 rType2 msg_fn
	in unify_state end

    | unify unify_state (t1 as ForallType(_)) t2 msg_fn =
        raise UnificationError("BUG: not expecting to unify forall type " ^
			       typeToString(t1) ^ " and " ^ 
			       typeToString(t2) ^ " at " ^ msg_fn())

    | unify unify_state t1 (t2 as ForallType(_)) msg_fn =
        unify unify_state t2 t1 msg_fn

    | unify unify_state t1 t2 msg_fn =
        raise UnificationError(typeToString(t1) ^ " doesn't match " ^ 
			       typeToString(t2) ^ " at " ^ msg_fn())


  fun instantiateType unify_state (ForallType{typevars,typ}) =
        let
	  (* make a type substitution, mapping each quantified type
	     variable to a fresh type var of the same kind (regular or
	     equality type) *)
	  val (subst:TypeSubstitution, unify_state) =
	    foldl
	      (fn(tv:string, (subst,unify_state))=>
                let val (fresh:Type, unify_state) =
		      fresh_type_var_of_kind unify_state (is_equality_type tv)
		in (add_type_var(subst, tv, fresh), unify_state) end)
	      (empty_type_subst, unify_state) typevars

	  (* then apply the substitution to the body of the forall type *)
	  val instantiatedType = apply_type_subst(subst, typ)

	in (instantiatedType, unify_state) end

    | instantiateType unify_state typ =
        (* if not a quantified type, then do nothing *)
        (typ, unify_state)


  fun generalizeTypeEnv unify_state tenv enclosingTEnv =
    let
      (* first compute the set of type variables free in the enclosing
	 type environment, after resolving it; these are
	 the non-generalizable type variables *)
      val nonGenericVars:string list =
        Assoc_List.iterate(
          (fn((name:string,typ:Type), prevNonGenericVars:string list) =>
	     let val resolvedTyp = resolve unify_state typ
		 val freeVars = getFreeVars resolvedTyp
	     in union(prevNonGenericVars, freeVars) end),
          nil, enclosingTEnv)

      (* helper to process a single type *)
      fun generalizeType(typ:Type):Type =
        let
          (* resolve the type *)
	  val resolvedTyp = resolve unify_state typ

	  (* get its free vars *)
	  val freeVars = getFreeVars(resolvedTyp)

	  (* remove any non-generalizable free vars *)
          val genericFreeVars =
	    List.filter
	      (fn(fv:string) => not(isMember fv nonGenericVars))
	      freeVars
        in
          (* introduce a forall type if there are any generalizable
	     free vars *)
          if genericFreeVars <> nil then
	    ForallType{typevars=genericFreeVars, typ=resolvedTyp}
	  else
	    resolvedTyp
        end

    in
      (* iterate over each binding, generalizing its free types, after
	 resolution of constraints *)
      Assoc_List.iterate((fn((name,typ), currTenv) =>
		               add_binding(currTenv, name,
					   generalizeType(typ))),
		         empty_environment, tenv)
    end


  fun toStringIndenting {map,type_var_counter} indent =
    (* for now, just print out the mapping *)
    Assoc_List.iterate(
      (fn((name:string, typ:Type), prefix:string) =>
        prefix ^ indent ^ name ^ " => " ^ typeToString(typ) ^ "\n"), "", map)

  fun toString unify_state = toStringIndenting unify_state ""

end

end
